@@ -4,7 +4,7 @@ import pytest
 
 from bleach import clean
 from bleach.html5lib_shim import Filter
-from bleach.sanitizer import ALLOWED_PROTOCOLS, Cleaner
+from bleach.sanitizer import ALLOWED_PROTOCOLS, Cleaner, NoCssSanitizerWarning
 from bleach._vendor.html5lib.constants import rcdataElements
 
 
@@ -29,7 +29,7 @@ def test_clean_idempotent(data):
 
 
 def test_clean_idempotent_img():
-    tags = ["img"]
+    tags = {"img"}
     dirty = '<imr src="http://example.com?foo=bar&bar=foo&amp;biz=bash">'
     assert clean(clean(dirty, tags=tags), tags=tags) == clean(dirty, tags=tags)
 
@@ -156,6 +156,22 @@ def test_bare_entities_get_escaped_correctly(text, expected):
     assert clean(text) == expected
 
 
+@pytest.mark.parametrize(
+    "text, expected",
+    [
+        ("x<y", "x&lt;y"),
+        ("<y", "&lt;y"),
+        ("x < y", "x &lt; y"),
+        ("<y>", "&lt;y&gt;"),
+    ],
+)
+def test_lessthan_escaping(text, expected):
+    # Tests whether < gets escaped correctly in a series of edge cases where
+    # the html5lib tokenizer hits an error because it's not the beginning of a
+    # tag.
+    assert clean(text) == expected
+
+
 @pytest.mark.parametrize(
     "text, expected",
     [
@@ -238,21 +254,21 @@ def test_character_entities_handling(text, expected):
         # a tag is disallowed, so it's stripped
         (
             '<p><a href="http://example.com/">link text</a></p>',
-            {"tags": ["p"]},
+            {"tags": {"p"}},
             "<p>link text</p>",
         ),
         # Test nested disallowed tag
         (
             "<p><span>multiply <span>nested <span>text</span></span></span></p>",
-            {"tags": ["p"]},
+            {"tags": {"p"}},
             "<p>multiply nested text</p>",
         ),
         # (#271)
-        ("<ul><li><script></li></ul>", {"tags": ["ul", "li"]}, "<ul><li></li></ul>"),
+        ("<ul><li><script></li></ul>", {"tags": {"ul", "li"}}, "<ul><li></li></ul>"),
         # Test disallowed tag that's deep in the tree
         (
             '<p><a href="http://example.com/"><img src="http://example.com/"></a></p>',
-            {"tags": ["a", "p"]},
+            {"tags": {"a", "p"}},
             '<p><a href="http://example.com/"></a></p>',
         ),
         # Test isindex -- the parser expands this to a prompt (#279)
@@ -272,11 +288,8 @@ def test_character_entities_handling(text, expected):
 )
 def test_stripping_tags(data, kwargs, expected):
     assert clean(data, strip=True, **kwargs) == expected
-    assert clean("  " + data + "  ", strip=True, **kwargs) == "  " + expected + "  "
-    assert (
-        clean("abc " + data + " def", strip=True, **kwargs)
-        == "abc " + expected + " def"
-    )
+    assert clean(f"  {data}  ", strip=True, **kwargs) == f"  {expected}  "
+    assert clean(f"abc {data} def", strip=True, **kwargs) == f"abc {expected} def"
 
 
 @pytest.mark.parametrize(
@@ -308,8 +321,8 @@ def test_stripping_tags(data, kwargs, expected):
 )
 def test_escaping_tags(data, expected):
     assert clean(data, strip=False) == expected
-    assert clean("  " + data + "  ", strip=False) == "  " + expected + "  "
-    assert clean("abc " + data + " def", strip=False) == "abc " + expected + " def"
+    assert clean(f"  {data}  ", strip=False) == f"  {expected}  "
+    assert clean(f"abc {data} def", strip=False) == f"abc {expected} def"
 
 
 @pytest.mark.parametrize(
@@ -329,9 +342,7 @@ def test_href_with_wrong_tag():
 
 
 def test_disallowed_attr():
-    IMG = [
-        "img",
-    ]
+    IMG = {"img"}
     IMG_ATTR = ["src"]
 
     assert clean('<a onclick="evil" href="test">test</a>') == '<a href="test">test</a>'
@@ -360,9 +371,7 @@ def test_unquoted_event_handler_attr_value():
 
 
 def test_invalid_filter_attr():
-    IMG = [
-        "img",
-    ]
+    IMG = {"img"}
     IMG_ATTR = {
         "img": lambda tag, name, val: name == "src" and val == "http://example.com/"
     }
@@ -387,7 +396,7 @@ def test_invalid_filter_attr():
 
 def test_poster_attribute():
     """Poster attributes should not allow javascript."""
-    tags = ["video"]
+    tags = {"video"}
     attrs = {"video": ["poster"]}
 
     test = '<video poster="javascript:alert(1)"></video>'
@@ -400,7 +409,7 @@ def test_poster_attribute():
 def test_attributes_callable():
     """Verify attributes can take a callable"""
     ATTRS = lambda tag, name, val: name == "title"
-    TAGS = ["a"]
+    TAGS = {"a"}
 
     text = '<a href="/foo" title="blah">example</a>'
     assert clean(text, tags=TAGS, attributes=ATTRS) == '<a title="blah">example</a>'
@@ -412,7 +421,7 @@ def test_attributes_wildcard():
         "*": ["id"],
         "img": ["src"],
     }
-    TAGS = ["img", "em"]
+    TAGS = {"img", "em"}
 
     text = (
         'both <em id="foo" style="color: black">can</em> have <img id="bar" src="foo"/>'
@@ -426,7 +435,7 @@ def test_attributes_wildcard():
 def test_attributes_wildcard_callable():
     """Verify attributes[*] callable works"""
     ATTRS = {"*": lambda tag, name, val: name == "title"}
-    TAGS = ["a"]
+    TAGS = {"a"}
 
     assert (
         clean('<a href="/foo" title="blah">example</a>', tags=TAGS, attributes=ATTRS)
@@ -443,7 +452,7 @@ def test_attributes_tag_callable():
     ATTRS = {
         "img": img_test,
     }
-    TAGS = ["img"]
+    TAGS = {"img"}
 
     text = 'foo <img src="http://example.com" alt="blah"> baz'
     assert clean(text, tags=TAGS, attributes=ATTRS) == "foo <img> baz"
@@ -457,7 +466,7 @@ def test_attributes_tag_callable():
 def test_attributes_tag_list():
     """Verify attributes[tag] list works"""
     ATTRS = {"a": ["title"]}
-    TAGS = ["a"]
+    TAGS = {"a"}
 
     assert (
         clean('<a href="/foo" title="blah">example</a>', tags=TAGS, attributes=ATTRS)
@@ -468,7 +477,7 @@ def test_attributes_tag_list():
 def test_attributes_list():
     """Verify attributes list works"""
     ATTRS = ["title"]
-    TAGS = ["a"]
+    TAGS = {"a"}
 
     text = '<a href="/foo" title="blah">example</a>'
     assert clean(text, tags=TAGS, attributes=ATTRS) == '<a title="blah">example</a>'
@@ -505,78 +514,83 @@ def test_attributes_list():
         # Specified protocols are allowed
         (
             '<a href="myprotocol://more_text">allowed href</a>',
-            {"protocols": ["myprotocol"]},
+            {"protocols": {"myprotocol"}},
             '<a href="myprotocol://more_text">allowed href</a>',
         ),
         # Unspecified protocols are not allowed
         (
             '<a href="http://example.com">invalid href</a>',
-            {"protocols": ["myprotocol"]},
+            {"protocols": {"myprotocol"}},
             "<a>invalid href</a>",
         ),
         # Anchors are ok
         (
             '<a href="#section-1">foo</a>',
-            {"protocols": []},
+            {"protocols": set()},
             '<a href="#section-1">foo</a>',
         ),
         # Anchor that looks like a domain is ok
         (
             '<a href="#example.com">foo</a>',
-            {"protocols": []},
+            {"protocols": set()},
             '<a href="#example.com">foo</a>',
         ),
-        # Allow implicit http if allowed
+        # Allow implicit http/https if allowed
         (
             '<a href="/path">valid</a>',
-            {"protocols": ["http"]},
+            {"protocols": {"http"}},
+            '<a href="/path">valid</a>',
+        ),
+        (
+            '<a href="/path">valid</a>',
+            {"protocols": {"https"}},
             '<a href="/path">valid</a>',
         ),
         (
             '<a href="example.com">valid</a>',
-            {"protocols": ["http"]},
+            {"protocols": {"http"}},
             '<a href="example.com">valid</a>',
         ),
         (
             '<a href="example.com:8000">valid</a>',
-            {"protocols": ["http"]},
+            {"protocols": {"http"}},
             '<a href="example.com:8000">valid</a>',
         ),
         (
             '<a href="localhost">valid</a>',
-            {"protocols": ["http"]},
+            {"protocols": {"http"}},
             '<a href="localhost">valid</a>',
         ),
         (
             '<a href="localhost:8000">valid</a>',
-            {"protocols": ["http"]},
+            {"protocols": {"http"}},
             '<a href="localhost:8000">valid</a>',
         ),
         (
             '<a href="192.168.100.100">valid</a>',
-            {"protocols": ["http"]},
+            {"protocols": {"http"}},
             '<a href="192.168.100.100">valid</a>',
         ),
         (
             '<a href="192.168.100.100:8000">valid</a>',
-            {"protocols": ["http"]},
+            {"protocols": {"http"}},
             '<a href="192.168.100.100:8000">valid</a>',
         ),
         pytest.param(
             *(
                 '<a href="192.168.100.100:8000/foo#bar">valid</a>',
-                {"protocols": ["http"]},
+                {"protocols": {"http"}},
                 '<a href="192.168.100.100:8000/foo#bar">valid</a>',
             ),
             marks=pytest.mark.xfail,
         ),
-        # Disallow implicit http if disallowed
-        ('<a href="example.com">foo</a>', {"protocols": []}, "<a>foo</a>"),
-        ('<a href="example.com:8000">foo</a>', {"protocols": []}, "<a>foo</a>"),
-        ('<a href="localhost">foo</a>', {"protocols": []}, "<a>foo</a>"),
-        ('<a href="localhost:8000">foo</a>', {"protocols": []}, "<a>foo</a>"),
-        ('<a href="192.168.100.100">foo</a>', {"protocols": []}, "<a>foo</a>"),
-        ('<a href="192.168.100.100:8000">foo</a>', {"protocols": []}, "<a>foo</a>"),
+        # Disallow implicit http/https if disallowed
+        ('<a href="example.com">foo</a>', {"protocols": set()}, "<a>foo</a>"),
+        ('<a href="example.com:8000">foo</a>', {"protocols": set()}, "<a>foo</a>"),
+        ('<a href="localhost">foo</a>', {"protocols": set()}, "<a>foo</a>"),
+        ('<a href="localhost:8000">foo</a>', {"protocols": set()}, "<a>foo</a>"),
+        ('<a href="192.168.100.100">foo</a>', {"protocols": set()}, "<a>foo</a>"),
+        ('<a href="192.168.100.100:8000">foo</a>', {"protocols": set()}, "<a>foo</a>"),
         # Disallowed protocols with sneaky character entities
         ('<a href="javas&#x09;cript:alert(1)">alert</a>', {}, "<a>alert</a>"),
         ('<a href="&#14;javascript:alert(1)">alert</a>', {}, "<a>alert</a>"),
@@ -595,7 +609,7 @@ def test_uri_value_allowed_protocols(data, kwargs, expected):
 def test_svg_attr_val_allows_ref():
     """Unescape values in svg attrs that allow url references"""
     # Local IRI, so keep it
-    TAGS = ["svg", "rect"]
+    TAGS = {"svg", "rect"}
     ATTRS = {
         "rect": ["fill"],
     }
@@ -607,7 +621,7 @@ def test_svg_attr_val_allows_ref():
     )
 
     # Non-local IRI, so drop it
-    TAGS = ["svg", "rect"]
+    TAGS = {"svg", "rect"}
     ATTRS = {
         "rect": ["fill"],
     }
@@ -631,7 +645,7 @@ def test_svg_attr_val_allows_ref():
 )
 def test_svg_allow_local_href(text, expected):
     """Keep local hrefs for svg elements"""
-    TAGS = ["svg", "pattern"]
+    TAGS = {"svg", "pattern"}
     ATTRS = {
         "pattern": ["id", "href"],
     }
@@ -653,7 +667,7 @@ def test_svg_allow_local_href(text, expected):
 )
 def test_svg_allow_local_href_nonlocal(text, expected):
     """Drop non-local hrefs for svg elements"""
-    TAGS = ["svg", "pattern"]
+    TAGS = {"svg", "pattern"}
     ATTRS = {
         "pattern": ["id", "href"],
     }
@@ -723,7 +737,7 @@ def test_nonexistent_namespace():
     ],
 )
 def test_self_closing_tags_self_close(tag):
-    assert clean("<%s>" % tag, tags=[tag]) == "<%s>" % tag
+    assert clean(f"<{tag}>", tags={tag}) == f"<{tag}>"
 
 
 # tags that get content passed through (i.e. parsed with parseRCDataRawtext)
@@ -744,16 +758,15 @@ _raw_tags = [
     [
         (
             raw_tag,
-            "<noscript><%s></noscript><img src=x onerror=alert(1) />" % raw_tag,
-            "<noscript>&lt;%s&gt;</noscript>&lt;img src=x onerror=alert(1) /&gt;"
-            % raw_tag,
+            f"<noscript><{raw_tag}></noscript><img src=x onerror=alert(1) />",
+            f"<noscript>&lt;{raw_tag}&gt;</noscript>&lt;img src=x onerror=alert(1) /&gt;",
         )
         for raw_tag in _raw_tags
     ],
 )
 def test_noscript_rawtag_(raw_tag, data, expected):
     # refs: bug 1615315 / GHSA-q65m-pv3f-wr5r
-    assert clean(data, tags=["noscript", raw_tag]) == expected
+    assert clean(data, tags={"noscript", raw_tag}) == expected
 
 
 @pytest.mark.parametrize(
@@ -762,10 +775,15 @@ def test_noscript_rawtag_(raw_tag, data, expected):
         (
             namespace_tag,
             rc_data_element_tag,
-            "<%s><%s><img src=x onerror=alert(1)>"
-            % (namespace_tag, rc_data_element_tag),
-            "<%s><%s>&lt;img src=x onerror=alert(1)&gt;</%s></%s>"
-            % (namespace_tag, rc_data_element_tag, rc_data_element_tag, namespace_tag),
+            (
+                f"<{namespace_tag}><{rc_data_element_tag}>"
+                + "<img src=x onerror=alert(1)>"
+            ),
+            (
+                f"<{namespace_tag}><{rc_data_element_tag}>"
+                + "&lt;img src=x onerror=alert(1)&gt;"
+                + f"</{rc_data_element_tag}></{namespace_tag}>"
+            ),
         )
         for namespace_tag in ["math", "svg"]
         # https://dev.w3.org/html5/html-author/#rcdata-elements
@@ -781,7 +799,7 @@ def test_namespace_rc_data_element_strip_false(
     #
     # browsers will pull the img out of the namespace and rc data tag resulting in XSS
     assert (
-        clean(data, tags=[namespace_tag, rc_data_element_tag], strip=False) == expected
+        clean(data, tags={namespace_tag, rc_data_element_tag}, strip=False) == expected
     )
 
 
@@ -1065,7 +1083,7 @@ def test_html_comments_escaped(namespace_tag, end_tag, eject_tag, data, expected
     #
     # the ejected elements can trigger XSS
     assert (
-        clean(data, tags=[namespace_tag, end_tag, eject_tag], strip_comments=False)
+        clean(data, tags={namespace_tag, end_tag, eject_tag}, strip_comments=False)
         == expected
     )
 
@@ -1092,7 +1110,7 @@ def test_html_comments_escaped(namespace_tag, end_tag, eject_tag, data, expected
             # newlines are used to keep lists and other elements readable
             (
                 "<div><p>This is our <b>description!</b> &amp;</p><p>1</p>"
-                "<ul><li>a</li><li>b</li><li>c</li></ul></div><p>nice!</p>"
+                + "<ul><li>a</li><li>b</li><li>c</li></ul></div><p>nice!</p>"
             ),
             "\nThis is our description! &amp;\n1\n\na\nb\nc\nnice!",
         ),
@@ -1103,7 +1121,7 @@ def test_strip_respects_block_level_elements(text, expected):
     Insert a newline between block level elements
     https://github.com/mozilla/bleach/issues/369
     """
-    assert clean(text, tags=[], strip=True) == expected
+    assert clean(text, tags=set(), strip=True) == expected
 
 
 def get_ids_and_tests():
@@ -1149,14 +1167,28 @@ def test_regressions(test_case):
 
 def test_preserves_attributes_order():
     html = """<a target="_blank" href="https://example.com">Link</a>"""
-    cleaned_html = clean(html, tags=["a"], attributes={"a": ["href", "target"]})
+    cleaned_html = clean(html, tags={"a"}, attributes={"a": ["href", "target"]})
 
     assert cleaned_html == html
 
 
+@pytest.mark.parametrize(
+    "attr",
+    (
+        ["style"],
+        {"*": ["style"]},
+    ),
+)
+def test_css_sanitizer_warning(attr):
+    # If you have "style" in attributes, but don't set a css_sanitizer, it
+    # should raise a warning.
+    with pytest.warns(NoCssSanitizerWarning):
+        clean("foo", attributes=attr)
+
+
 class TestCleaner:
     def test_basics(self):
-        TAGS = ["span", "br"]
+        TAGS = {"span", "br"}
         ATTRS = {"span": ["style"]}
 
         cleaner = Cleaner(tags=TAGS, attributes=ATTRS)
@@ -1178,7 +1210,7 @@ class TestCleaner:
                     yield token
 
         ATTRS = {"img": ["rel", "src"]}
-        TAGS = ["img"]
+        TAGS = {"img"}
 
         cleaner = Cleaner(tags=TAGS, attributes=ATTRS, filters=[MooFilter])
 
